from sklearn.tree import DecisionTreeClassifier
from sklearn import svm
from sklearn.linear_model import SGDClassifier
from sklearn.neighbors import KNeighborsClassifier
from sklearn.ensemble import GradientBoostingClassifier
from sklearn.ensemble import AdaBoostClassifier
from sklearn.ensemble import RandomForestClassifier
from sklearn.neural_network import MLPClassifier
from sklearn.metrics import accuracy_score
from sklearn.ensemble import ExtraTreesClassifier

def default_modeling(df, y):
    svm_model = svm.SVC().fit(df['train'], y['train'])
    neighbors = KNeighborsClassifier().fit(df['train'], y['train'])
    tree = DecisionTreeClassifier().fit(df['train'], y['train'])
    sgd = SGDClassifier().fit(df['train'], y['train'])
    grad = GradientBoostingClassifier().fit(df['train'], y['train'])
    adaboost = AdaBoostClassifier().fit(df['train'], y['train'])
    forest = RandomForestClassifier().fit(df['train'], y['train'])
    network = MLPClassifier().fit(df['train'], y['train'])

    list_models = [svm_model, neighbors, tree, sgd, grad, adaboost, forest, network]
    scores = []

    for model in list_models:
        scores.append(accuracy_score(model.predict(df['test']), y['test']))

    return list_models[scores.index(max(scores))]

def a_svm(df_1, y, trained=True):
    clf = svm.SVC()
    if trained:
        clf.fit(df_1, y)

    return clf

def nearest_neighbors(df_1, y, trained=True):
    neigh = KNeighborsClassifier()

    min_neighbors = 3
    max_neighbors = 13
    best_score = 0
    best_model = None
    best_neighbors = 0

    sample_amount = int(len(df_1) * 0.5)
    sampled_df = df_1.iloc[:sample_amount]
    sampled_y = y.iloc[:sample_amount]


    if trained:
        for x in range(min_neighbors, max_neighbors):
            neigh = KNeighborsClassifier(n_neighbors=x)
            neigh.fit(sampled_df, sampled_y)
            a_score = accuracy_score(neigh.predict(sampled_df), sampled_y)
            if a_score > best_score:
                best_score=a_score
                best_neighbors=x
                best_model=neigh

        model = KNeighborsClassifier(n_neighbors=best_neighbors).fit(df_1, y)
        
    return model


def a_tree(df_1, y, trained=True):
    tree = DecisionTreeClassifier()
    if trained:
        tree.fit(df_1, y)

    return tree


def sgd(df_1, y, trained=True):
    clf = SGDClassifier()
    if trained:
        clf.fit(df_1, y)

    return clf


def gradient_boosting(df_1, y, trained=True):
    clf = GradientBoostingClassifier()
    if trained:
        clf.fit(df_1, y)

    return clf


def adaboost(df_1, y, trained=True):
    clf = AdaBoostClassifier()
    if trained:
        clf.fit(df_1, y)
    return clf


def rf(df_1, y, trained=True):
    clf = RandomForestClassifier()
    if trained:
        clf.fit(df_1, y)
    return clf

def mlp(df_1, y, trained=True):
    clf = MLPClassifier()
    if trained:
        clf.fit(df_1, y)
    return clf

def svm_stroke(df_1 , y):
    models = []
    scores = []

    clf_1 = svm.SVC().fit(df_1, y)
    clf_2 = svm.SVC(kernel='rbf').fit(df_1, y)
    clf_3 = svm.SVC(kernel='poly').fit(df_1, y)

    models.append(clf_1)
    models.append(clf_2)
    models.append(clf_3)

    for model in models:
        scores.append(accuracy_score(model.predict(df_1), y))

    return models[scores.index(max(scores))]

def ensemble_stroke(df_1, y):
    models = []
    scores = []

    grad = GradientBoostingClassifier().fit(df_1, y)
    adaboost = AdaBoostClassifier().fit(df_1, y)
    forest = RandomForestClassifier().fit(df_1, y)
    extra = ExtraTreesClassifier().fit(df_1, y)

    models.append(grad)
    models.append(adaboost)
    models.append(forest)
    models.append(extra)

    for model in models:
        scores.append(accuracy_score(model.predict(df_1), y))

    return models[scores.index(max(scores))]






import numpy as np
import pandas as pd
from sklearn.compose import ColumnTransformer
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.pipeline import Pipeline
from prince.ca import CA
from nltk.stem import WordNetLemmatizer
from nltk.tokenize import word_tokenize
from nltk.corpus import stopwords
from autocorrect import Speller
from sklearn.model_selection import train_test_split
from sklearn.impute import SimpleImputer
from sklearn.preprocessing import (OneHotEncoder,
                                   StandardScaler,
                                   FunctionTransformer, LabelEncoder)


def initial_preprocessor(
        data,
        json_file):
    object_columns = [
        col for col,
                col_type in data.dtypes.iteritems() if col_type == 'object']

    # Handles dates without timestamps
    for col in object_columns:
        try:
            data[col] = pd.to_datetime(data[col], infer_datetime_format=True)
        except ValueError:
            pass

    # get target column
    target = json_file['data']['target']
    y = data[target]

    # remove rows where target is NaN
    data = data[y.notna()]
    y = y[y.notna()]

    del data[target]
    X_train, X_test, y_train, y_test = train_test_split(
        data, y, test_size=0.2)

    data = {
        'train': pd.concat([X_train], axis=1),
        'test': pd.concat([X_test], axis=1)
    }
    # preprocess the dataset
    full_pipeline = None
    if True:
        data, full_pipeline = structured_preprocessor(data)
    else:
        data.fillna(0, inplace=True)

    y = {'train': y_train, 'test': y_test}

    return data, y


def structured_preprocessor(data, ca_threshold=0.5, text=[]):
    # Preprocessing for datetime columns
    process_dates(data)

    # This will be used inside process_text once complete
    if len(text) > 0:
        text_preprocessing(data, text)

    # identifies the categorical and numerical columns
    categorical_columns = data['train'].select_dtypes(
        exclude=["number"]).columns
    numeric_columns = data['train'].columns[data['train'].dtypes.apply(
        lambda c: np.issubdtype(c, np.number))]

    # Removes text columns from categorical columns to use in separate pipeline
    categorical_columns = [
        cat_cols for cat_cols in categorical_columns if cat_cols not in text]

    full_pipeline = ColumnTransformer([], remainder="passthrough")

    if len(numeric_columns) != 0:
        # pipeline for numeric columns
        num_pipeline = Pipeline([
            ('imputer', SimpleImputer(strategy="median")),
            ('std_scaler', StandardScaler())
        ])

        full_pipeline.transformers.append(
            ("num", num_pipeline, numeric_columns))

    if len(text) != 0:
        # Each text col needs a separate pipeline
        for x in range(len(text)):
            full_pipeline.transformers.append(
                (f"text_{x}",
                 Pipeline(
                     [
                         ('test',
                          FunctionTransformer(
                              lambda x: np.reshape(
                                  x.to_numpy(),
                                  (-1,
                                   1)))),
                         ('imputer',
                          SimpleImputer(
                              strategy="constant",
                              fill_value="")),
                         ('raveler',
                          FunctionTransformer(
                              lambda x: x.ravel(),
                              accept_sparse=True)),
                         ('vect',
                          TfidfVectorizer()),
                         ('densifier',
                          FunctionTransformer(
                              lambda x: x.todense(),
                              accept_sparse=True)),
                         ('embedder',
                          FunctionTransformer(
                              textembedder,
                              accept_sparse=True))]),
                 text[x]))

    if len(categorical_columns) != 0:
        combined = pd.concat([data['train'], data['test']], axis=0)

        ca_threshold = combined.shape[0] * \
                       .25 if ca_threshold is None else combined.shape[0] * ca_threshold

        if too_many_values(combined[categorical_columns], ca_threshold):
            cat_pipeline = Pipeline([
                ('imputer', SimpleImputer(strategy="constant", fill_value="")),
                ('one_hotencoder', OneHotEncoder(handle_unknown='ignore')),
                ('transformer', FunctionTransformer(lambda x: x.toarray(), accept_sparse=True)),
                ('ca', CA(n_components=-1))
            ])
        else:
            cat_pipeline = Pipeline([
                ('imputer', SimpleImputer(strategy="constant", fill_value="")),
                ('one_hotencoder', OneHotEncoder(handle_unknown='ignore'))
            ])

        full_pipeline.transformers.append(
            ('cat', cat_pipeline, categorical_columns))

    train = full_pipeline.fit_transform(data['train'])

    train_cols = generate_column_labels(full_pipeline, numeric_columns, text)

    test = full_pipeline.transform(data['test'])

    test_cols = generate_column_labels(full_pipeline, numeric_columns, text)

    # Ternary clause because when running housing.csv,
    # the product of preprocessing is np array, but not when using landslide
    # data... not sure why
    data['train'] = pd.DataFrame(
        (train.toarray() if not isinstance(
            train,
            np.ndarray) else train),
        columns=train_cols)
    data['test'] = pd.DataFrame(
        (test.toarray() if not isinstance(
            test,
            np.ndarray) else test),
        columns=test_cols)
    return data, full_pipeline


def process_dates(data):

    for df in data:
        df = data[df]
        datetime_cols = df.select_dtypes('datetime64')
        for col in datetime_cols:
            if col in columns:
                df[f'{col}_DayOfWeek'] = df[col].dt.day_name()
                df[f'{col}_Year'] = df[col].dt.year
                df[f'{col}_Month'] = df[col].dt.month_name()
                df[f'{col}_MonthDay'] = df[col].dt.day

                del df[col]


# Preprocesses text for word embedding


def text_preprocessing(data, text_cols):
    lemmatizer = WordNetLemmatizer()
    combined = pd.concat([data['train'], data['test']], axis=0)

    spell = Speller(fast=True)
    for col in text_cols:
        combined[col] = combined[col].apply(
            lambda x: x.lower() if isinstance(x, str) else x)

    stop_words = set(stopwords.words('english'))

    for col in text_cols:
        preprocessed_text = []
        for words in combined[col]:
            if words is not np.nan:
                words = word_tokenize(words)
                words = [word for word in words if word.isalpha()]
                words = [word for word in words if word not in stop_words]
                words = [spell(word) for word in words]
                words = [lemmatizer.lemmatize(word) for word in words]

                preprocessed_text.append(' '.join(words))

            else:
                preprocessed_text.append(np.nan)

        combined[col] = preprocessed_text
    data['train'] = combined.iloc[:len(data['train'])]
    data['test'] = combined.iloc[len(data['train']):]


def textembedder(text):
    total = list()
    for i in text:
        total.append(np.sum(i))

    return np.reshape(total, (-1, 1))


# Sees if one hot encoding occurred, if not just uses numeric cols


def generate_column_labels(full_pipeline, numeric_cols, text_cols):
    # Check if one hot encoding was performed
    if 'cat' in full_pipeline.named_transformers_:
        # If ca was used
        if isinstance(full_pipeline.named_transformers_['cat'][-1], CA):
            ca = full_pipeline.named_transformers_['cat'][-1]
            encoded_cols = [f'CA_{x}' for x in range(len(ca.eigenvalues_))]
            cols = [*list(numeric_cols), *encoded_cols, *text_cols]

        else:
            try:
                encoded_cols = full_pipeline.named_transformers_[
                    'cat']['one_hotencoder'].get_feature_names()
                cols = [*list(numeric_cols), *encoded_cols, *text_cols]

            except Exception as error:
                # For debugging only
                print(error)
                cols = list(numeric_cols)

        return cols
    else:
        return [*list(numeric_cols), *text_cols]


# Method to calculate how many columns the data set will
# have after one hot encoding
# Decides whether CA is needed or not essentially
# mca_threshold is the len of the dataset * .25 to calculate the proportion of
# when to apply CA
def too_many_values(data, ca_threshold):
    total_unique = 0
    for col in data:

        if total_unique > ca_threshold:
            return True
        # Use value_counts() due to same columns having strings and floats
        total_unique += len(data[col].value_counts())

    return False



import pandas as pd

class DataReader():
    def __init__(self, in_mem_data, dataset_name):
        self.data = in_mem_data
        self.dataset_name = dataset_name

    def data_reader(self):
        frame = 0
        if 'type' not in self.data:
            data_type = self.type_detector(self.dataset_name)
        else:
            data_type = self.data['type']

        dropping = []
        if "drop" not in self.data:
            pass
        else:
            for element in self.data["drop"]:
                if element == self.data["target"]:
                    raise Exception(
                        "Your target of --  {} -- cannot be the same as the columns you're trying to drop.".format(
                            self.data["target"]))
                dropping.append(element)

        if data_type == 'csv':
            frame = self.csv(self.dataset_name, drop=dropping)
            if self.data['data']['target'] not in frame.columns:
                raise Exception("The target you provided -- {} -- is not in the dataframe.".format(self.data["target"]))
        if data_type == 'xs':
            frame = self.xs(self.data, drop=dropping)
            if self.data['target'] not in frame.columns:
                raise Exception("The target you provided -- {} -- is not in the dataframe.".format(self.data["target"]))
        if 'trim' not in self.data:
            pass
        else:
            if self.data['trim'] < 0.1 or self.data['train'] > 0.9:
                raise Exception("Trimming amounts should be between the value 0.1 and 0.9.")
            target_size = int(self.data['trim'] * len(frame))

            while target_size > len(frame):
                frame = frame.iloc[1:]

        return frame

    def type_detector(self, data):
        if data[-3:] == 'csv':
            return 'csv'
        if data[-2:] == 'xs':
            return 'xs'
        if data[-4:] == 'json':
            return 'json'

    def csv(self, dataset_name, drop=[]):

        try:
            df = pd.read_csv(dataset_name)
        except:
            raise Exception("The file path you provided -- {} -- does not exist".format(str(self.data)))
        for element in drop:
            try:
                del df[element]
            except:
                raise Exception("Your column -- {} --  is not in the dataframe".format(element))
        return df

    def xs(self, datset_name, drop=[]):
        try:
            df = pd.read_excel(datset_name)
        except:
            raise Exception("The file path you provided -- {} -- does not exist".format(datset_name))
        for element in drop:
            try:
                del df[element]
            except:
                raise Exception("Your column -- {} --  is not in the dataframe".format(element))